msnfa.init.para = function(Y, g, q)
{
    n = nrow(Y)
    p = ncol(Y)
    mu = matrix(NA, p, g)
    B = array(NA, dim = c(p,q,g))
    D = matrix(NA, p, g)
    la = matrix(NA,q,g)
    nu = rep(NA,g)
    kmY = kmeans(Y,g)
    clus = kmY$cluster
    w = table(clus)/n
   for(i in 1:g)
        {
        ind.Y = Y[clus == i,]
        mu[,i] = colMeans(ind.Y)
        cent = t(ind.Y) - mu[,i]
        V = cent %*% t(cent) / nrow(ind.Y)
        eig.V = eigen(V)
        B[,,i] = eig.V$ve[,1:q] %*% diag(sqrt(eig.V$va[1:q]), q, q)
        D[,i] = diag(V - B[,,i] %*% t(B[,,i]))
        S = B[,,i] %*% t(B[,,i]) + diag(D[,i])
        S.inv = solve(S)
        Ga = S.inv %*% B[,,i]
        Om = diag(q) - t(Ga) %*% B[,,i]
        x=t(t(Ga)%*%cent) # score
        la[,i]=EmSkew(x, g=1, dist='msn', debug=F)$delta
        }
     nu=rep(10000,g)
    return(list(w=w,mu = mu,B =B, D =D,la=la,nu=nu))
}

gener.na = function(X, na.rate)
{
  #set.seed(1)
  n = nrow(X)
  p = ncol(X)
  num.na = floor(n*p*na.rate)
  keep.part = sample(p, n, replace = T) + p * 0:(n-1)
  na.posi = tabulate(sample((1:(n*p))[-keep.part], num.na), nbins = n*p)
  na.posi = matrix(na.posi, ncol = p, byrow = T)
  X[na.posi == 1] = NA
  return(X)
}

sq.mt = function(M)
{
  p = ncol(M)
  if(p == 1) M.sqrt = as.matrix(sqrt(M))
  else
      {
       eig.M = eigen(M)
       M.sqrt = eig.M$ve %*% diag(sqrt(eig.M$va)) %*% t(eig.M$ve)
      }
}

dMST = function(Y, mu, S, la, nu, log = F)
{
    require(mvtnorm)
    if(is.vector(Y) == T) Y = as.matrix(Y)
    if(is.matrix(S) == F) S = as.matrix(S)
    
    p = length(la)

    cent = t(Y) - mu

    om = S + la %*% t(la)
    eigom = eigen(om)
    omh.inv = eigom$ve %*% diag(1/sqrt(eigom$va), p) %*% t(eigom$ve)
    
    e = omh.inv %*% cent
    Aj = colSums(e^2)
    tmp = c(omh.inv %*% la)
    xi = colSums(tmp * e)
    si2 = 1 - sum(tmp^2)

    r0 = (nu + p) / (nu + Aj)
    A = xi / sqrt(si2)
    c0 = A * sqrt(r0) ; Tc0 = pt(c0, df = nu + p)
    
    logtp = lgamma((nu+p) / 2) - lgamma(nu / 2) - log(det(om)) / 2 - 
    p / 2 * log(pi*nu) - (nu + p) / 2 * log(1 + Aj / nu)
    if(log == F){den =2*exp(logtp)*Tc0}
    else {den = log(2) + logtp + log(Tc0)}
    return(den)
}


MSNFA.na.ECM=function(Y, init.para, true.clus, zero.mu=F, zero.la=F, tol, max.iter, per)
{
  begin = proc.time()[1]
#initial values of missing
  n = nrow(Y)
  p = ncol(Y)
  w = init.para$w
  mu = init.para$mu
  B = init.para$B
  D = init.para$D
  la = init.para$la
  nu = init.para$nu
  q = ncol(B)
  g = length(w)
  if(is.null(init.para))
  {
    mu = colMeans(Y, na.rm = T)
    B = matrix(runif(p*q),p,q)
    D = diag(rep(1, p))
  }  

  CMQ.nu.fn = function(Y, Z ,mu.ast, S, d, nu)
    {
      logsum = c()
      for(j in 1:num.na.class)
      {   
        O = O.list[[j]]
        ind = ind.list[[j]]  
        Y.pat = matrix(Y[ind,] , ncol = p)
        Y.o = Y.pat%*%t(O)
        OSO = O %*% S %*% t(O)       #Sigma.oo#
        mu.ast.o=as.vector(O%*%mu.ast) #mu.o-cd.o#
        d.o = O%*%d  #do#
        logsum[j] = sum(Z[ind]*dMST(Y.o, mu.ast.o, OSO, d.o, nu,log = T))
      }
      -sum(logsum)
    }
    
    CML.nu.fn  = function(Y, Z ,mu.ast, S, d, nu)
    {
      wden = matrix(NA,n,g)
      for(i in 1:g)
      {
        for(j in 1:num.na.class)
        {   
          O = O.list[[j]]
          ind = ind.list[[j]]  
          Y.pat = matrix(Y[ind,] , ncol = p)
          Y.o = Y.pat%*%t(O)
          OSO = O %*% S[,,i] %*% t(O)       #Sigma.oo#
          mu.ast.o = as.vector(O%*%mu.ast[,i]) #mu.o-cd.o#
          d.o = O%*%d[,i]  #do#
          wden[ind,i] = w[i]*dMST(Y.o, mu.ast.o, OSO, d.o, nu)
        }
      }
      indv.den = rowSums(wden)
      return(-sum(log(indv.den)))  
    }


# O matrices
  na.posi = is.na(Y)                                 
  na.class = colSums(t(na.posi) * 2 ^ (0:(p-1)))     #ind.na
  uni.na.class = unique(na.class)                    #uni.ind
  num.na.class = length(uni.na.class)                #num.pat.na
  Ip = diag(p)
  O.list = ind.list = as.list(num.na.class)                                                                      
  for(i in 1:num.na.class)
  {
    ind.list[[i]]= which(na.class == uni.na.class[i])
    O.list[[i]] = matrix(diag(p)[!na.posi[ind.list[[i]][1],],], ncol = p)
  }
 Y[is.na(Y)] = 99999


#observed log-likelihood
    d = mu.ast = matrix(NA,p,g)
    S= array(NA, dim = c(p,p,g))
    wden = matrix(NA,n,g)
    if (zero.mu)  mu = matrix(0,p,g)
    if (zero.la)  la = matrix(0,q,g)
        
     for(i in 1:g)
    {
    anu = sqrt(nu[i] / pi) * exp(lgamma((nu[i]-1) / 2) - lgamma(nu[i]/2))
    Delta = diag(q) + (1 - (nu[i] - 2) / nu[i] * anu^2) * la[,i] %*% t(la[,i])
    eigDelta=eigen(Delta)
    Dh.inv = eigDelta$ve %*% diag(1/sqrt(eigDelta$va), q) %*% t(eigDelta$ve)
    B[,,i] = B[,,i]%*%Dh.inv
    S[,,i] = B[,,i] %*% t(B[,,i]) + diag(D[,i])
    d[,i] = c(as.matrix(B[,,i]) %*%la[,i])
    mu.ast[,i] = mu[,i] - anu * d[,i]
     for(j in 1:num.na.class)
     {   
     O = O.list[[j]]
     ind = ind.list[[j]]  
     Y.pat = matrix(Y[ind,] , ncol = p)
     Y.o = Y.pat%*%t(O)
     OSO = O %*% S[,,i] %*% t(O)       #Sigma.oo#
     mu.ast.o = as.vector(O%*%mu.ast[,i]) #mu.o-cd.o#
     d.o = O%*%d[,i]  #do#
     wden[ind,i] = w[i]*dMST(Y.o, mu.ast.o, OSO, d.o, nu[i]) 
     }
    }
    indv.den=rowSums(wden)
    init.LL=sum(log(indv.den))
      
   if (zero.la == F)
    {
        cat("\n\t\t  MSNFA.na\t ECME (g = ",g, " ; q= ", q, ")\t la_i~=0 for all i\n", sep = "")
    }
    else
    {
        cat("\n\t\t  MSNFA.na\t ECM (g= ",g, " ; q= ", q, ")\t la_i=0 for all i\n", sep = "")
    }
  iter = 1
  cat(paste(rep("=", 50), sep = "", collapse = ""), "\n")
  cat("iter = ", iter-1, "\t   init.LL = ",  init.LL, sep = "", "\n")
   LL=lk=init.LL
  epsilon=Inf

  while((iter < max.iter) && (epsilon > tol))
  {  
  Z=wden/indv.den
  ni=colSums(Z)
  w=ni/n
  hu= array(NA, dim = c(q,n,g))
  Y.hat=array(NA, dim = c(p,n,g))
#E-step   
  for(i in 1:g)
    {
    Reta=b.hat =DPS =matrix(NA, p, n)
    u1j=matrix(NA,q,n)
    tau=ga=g1j=g2j=matrix(NA,1,n)
    Phi.temp.Sum=matrix(0,p,p)
    RPsi.sum=matrix(0,p,q)
    hu2.sum=matrix(0,q,q)
    zeta1.sum=rep(0,q)
  #  anu = sqrt(nu[i] / pi) * exp(lgamma((nu[i]-1) / 2) - lgamma(nu[i]/2))
  #  S[,,i] = B[,,i] %*% t(B[,,i]) + diag(D[,i])
    om=S[,,i]+d[,i]%*%t(d[,i])   
    for(j in 1:num.na.class)
    {  
      ind = ind.list[[j]]
      O = O.list[[j]]
      po=nrow(O)
      OMO= O %*% om %*% t(O)
      ODO= O %*% diag(D[,i]) %*% t(O)
      Y.pat = matrix(Y[ind,] , ncol = p)
      Y.o=Y.pat%*%t(O)
      mu.ast.o=as.vector(O%*%mu.ast[,i])
      d.o=as.vector(O%*%d[,i])  
      mu.o=as.vector(O%*%mu[,i])
      
      OMoo=t(O) %*% solve(OMO) %*% O  #Soo# 
      Coo=t(O) %*% solve(ODO) %*% O #Coo#
      DCoo=diag(D[,i])%*%Coo
      IBoo=(diag(p)-DCoo)%*%B[,,i] #Roo
      IDoo=(diag(p)-DCoo)%*%diag(D[,i])
      
      cent.o =t(Y.o) - mu.ast.o
      eigom=eigen(OMO)
      om.sq.inv = eigom$ve %*% diag(1/sqrt(eigom$va), po) %*% t(eigom$ve)
      e = om.sq.inv %*% cent.o       #omegaoo^(-1/2)*(yo-mu.o-cdo)#
      Aj = colSums(e^2)
      tmp = c(om.sq.inv %*% d.o)   #omegaoo^(-1/2)* do#
      xi= colSums(tmp * e) 
      si2 = 1 - sum(tmp^2)
      A = xi / sqrt(si2)    
         
      rn2 = (nu[i] + po - 2) / (nu[i] + Aj); cn2 = A * sqrt(rn2)
      r0 = (nu[i] + po) / (nu[i] + Aj); c0 = A * sqrt(r0)
      r2 = (nu[i] + po + 2) / (nu[i] + Aj); c2 = A * sqrt(r2)
      Tc0 = pt(c0, df = nu[i] + po); Tc2 = pt(c2, df = nu[i] + po + 2)
      tc0 = dt(c0, df = nu[i] + po); tcn2 = dt(cn2, df = nu[i] + po - 2)
      
      tau[,ind] = r0 * Tc2 / Tc0
      g1j[,ind] = xi * tau[,ind] + sqrt(si2) * sqrt(r0) * tc0 / Tc0
      g2j[,ind] = si2 + xi * g1j[,ind]
      ga[,ind] = xi + sqrt(si2) * sqrt(rn2) * tcn2 / Tc0

      Woo = solve(diag(q) + t(B[,,i]) %*% Coo %*% B[,,i]) 
      v.o = t(B[,,i]) %*% Coo %*% (t(Y.pat)-mu[,i])
      
      hu[,ind,i] = Woo %*% (v.o + (la[,i]) %*% t(ga[,ind] - anu))
      u1j[,ind] = Woo %*% (t(t(v.o) * tau[,ind]) +
      (la[,i]) %*% t(g1j[,ind] - anu * tau[,ind])) #tau*U
      u2j = Woo %*% (t(t(v.o) * g1j[,ind]) + (la[,i]) %*% t(g2j[,ind] - anu * g1j[,ind]))   
      
      if(q==1)
      zeta1 = rowSums(t(Z[ind,i]*t(u2j))) -  anu * rowSums(t(t(Z[ind,i]*t(u1j[,ind]))))
      else
      zeta1 = rowSums(t(Z[ind,i]*t(u2j))) -  anu * rowSums(t(Z[ind,i]*t(u1j[,ind])))
      zeta1.sum=zeta1.sum+zeta1
      if(q==1) 
      hu2=(zeta1%*%t(la[,i])+t(Z[ind,i]*(u1j[,ind]))%*%t(v.o)+ diag(sum(Z[ind,i]),q))%*%Woo
      else 
      hu2=(zeta1%*%t(la[,i])+t(Z[ind,i]*t(u1j[,ind]))%*%t(v.o)+ diag(sum(Z[ind,i]),q))%*%Woo
      hu2.sum=hu2.sum+hu2
      RPsi.temp=IBoo%*%hu2     #A*om
      RPsi.sum=RPsi.sum+RPsi.temp
      Reta[,ind]= DCoo%*%B[,,i]%*%(u1j[,ind])
      b.hat[,ind] = mu[,i] + DCoo %*% (t(Y.pat)- mu[,i])
      if(q==1) 
      DPS[,ind]=IBoo%*%t(Z[ind,i]*(u1j[,ind])) 
      else 
      DPS[,ind]=IBoo%*%t(Z[ind,i]*t(u1j[,ind])) 
      if(q==1) 
      ze.hat = mu[,i] + B[,,i] %*% t(hu[,ind,i])
      else 
      ze.hat = mu[,i] + B[,,i] %*% (hu[,ind,i])
      Y.hat[,ind,i] = ze.hat + DCoo %*% (t(Y.pat)- ze.hat)
      Phi.temp =  sum(Z[ind,i]) * IDoo + IBoo%*%hu2%*%t(IBoo)
      Phi.temp.Sum = Phi.temp.Sum + Phi.temp  
    }  
    ##la
    zeta2 = sum(Z[,i]*t(g2j)) - 2 * anu * sum(Z[,i]*t(g1j)) +
    anu^2 * sum(Z[,i]*t(tau))
#M-step
    mu[,i]= colSums(c(Z[,i]*t(tau))*t(b.hat)- Z[,i]*t(Reta))/sum(Z[,i]*t(tau))
    B[,,i]=((b.hat - mu[,i])%*%(Z[,i]*t(u1j))+ RPsi.sum)%*% solve(hu2.sum)
    A.b = b.hat - mu[,i]
    ttmp=Phi.temp.Sum- RPsi.sum%*%t(B[,,i]) -B[,,i]%*% t(RPsi.sum)+
    B[,,i]%*% hu2.sum%*%t(B[,,i])
    Tmp=ttmp+A.b%*%t(DPS) -A.b%*%t(B[,,i]%*%t(Z[,i]*t(u1j)))+DPS%*%t(A.b)-
    B[,,i]%*%t(Z[,i]*t(u1j))%*%t(A.b)+(A.b)%*%(c(Z[,i]*t(tau))*t(A.b))
    D[,i] = diag(Tmp)/(ni[i])               
                              
   if (!zero.la) {
        la[,i]=zeta1.sum/zeta2
    }
    S[,,i] = B[,,i] %*% t(B[,,i]) + diag(D[,i])
    d[,i] = c(as.matrix(B[,,i]) %*%la[,i])
    mu.ast[,i] = mu[,i] - anu * d[,i]
   }
   
   for(i in 1:g)
   {
#observed log-likelihood  
    anu = sqrt(nu[i] / pi) * exp(lgamma((nu[i]-1) / 2) - lgamma(nu[i]/2))
  #  S[,,i] = B[,,i] %*% t(B[,,i]) + diag(D[,i])
  #  d[,i]=c(as.matrix(B[,,i]) %*%la[,i])
  #  mu.ast[,i]=mu[,i]-anu*d[,i]
  for(j in 1:num.na.class)
  {   
    O = O.list[[j]]
    OSO = O %*% S[,,i] %*% t(O)
    ind = ind.list[[j]]
   
    Y.pat = matrix(Y[ind,] , ncol = p)
    Y.o = Y.pat%*%t(O)
    mu.ast.o = as.vector(O%*%mu.ast[,i])
    d.o = O%*%d[,i]     
    wden[ind,i] = w[i]*dMST(Y.o, mu.ast.o, OSO, d.o, nu[i]) 
  }
    }
    indv.den=rowSums(wden)
    newLL =sum(log(indv.den))
    lk = c(lk, newLL)
    diff = newLL - LL
    ## Aikten's accerlation
            if (iter < 2) 
            {
                epsilon = abs(LL-newLL)/abs(newLL)
            }
            else
            {
                tmp = (newLL - LL)/(LL - lk[iter-1])
                tmp2 =  LL + (newLL - LL)/(1 - tmp)
                epsilon = abs(tmp2-newLL)
            }
            if(diff < 0)  cat('iter=', iter, "logL is decreasing!\n")
             LL =  newLL
            if (iter%%per == 0)  cat("iter =", iter, "\t newLL =", newLL, "\t diff =", diff,  "\t Aikten's diff =", epsilon, "\n")
            iter = iter + 1
  }
  cat(paste(rep("=", 60), sep = "", collapse = ""), "\n")
  cat("iter =", iter, "\t newLL =", newLL, "\t diff =", diff,  "\t Aikten's diff =", epsilon, "\n")
  Z=wden/indv.den
  post.clus = matrix(apply(Z,1,order),nrow=g)[g,]
  if(length(unique(post.clus))==length(unique(true.clus)))
  MCR=error.rate(post.clus, true.clus)
  else
  MCR=NA
  ARI=adjustedRandIndex(post.clus, true.clus)
  hu.sum = matrix(0,n,q)
  Y.sum = matrix(0,n,p)
  for(i in 1:g)
  {
    anu = sqrt(nu[i] / pi) * exp(lgamma((nu[i]-1) / 2) - lgamma(nu[i]/2))
    Delta = diag(q) + (1-(nu[i]-2)/nu[i]*anu^2) * la[,i] %*% t(la[,i])
    Dh=sq.mt(Delta)
    eigDelta=eigen(Delta)
    Dh.inv = eigDelta$ve %*% diag(1/sqrt(eigDelta$va), q) %*% t(eigDelta$ve)

    B[,,i] = B[,,i] %*% Dh
    hu[,,i] = Dh.inv %*% (hu[,,i])
    if(q==1)
    hu.sum=hu.sum+Z[,i]* (hu[,,i])
    else
    hu.sum=hu.sum+Z[,i]* t(hu[,,i])
    Y.sum=Y.sum+Z[,i]*t(Y.hat[,,i])
  }
    no.para = g*(2 * p + p * q - q * (q - 1)/2)+sum(la != 0)+g-1
    AIC = -2 * newLL +2* no.para
    BIC = -2 * newLL+ no.para * log(n)
    #abic2 = -2 * newLL + no.para * log(n/2+12)
    #ABIC= -abic2/2
    ENT = - sum(Z*log(Z+1e-10))
    ICL = BIC + 2* ENT
    AWE = -2 * newLL+ 2* ENT+ 2* no.para*(1.5 + log(n))
    model.inf = c(newLL = newLL, no.para=no.para,AIC=AIC,BIC=BIC,ICL=ICL,AWE=AWE, ARI=ARI, CCR=1-MCR)
    para = list(w=w,mu = mu, B = B, D = D, la = la,nu=nu)
    end = proc.time()[1]
    cat('no.para =', no.para, ' AIC =', AIC, ' BIC =', BIC, ' ICL=', ICL,' AWE=', AWE, ' CCR=', 1-MCR, ' ARI=', ARI, '\n', "MSNFA.na.ECM takes", end - begin, "seconds\n\n")
    cat(paste(rep("-", 60), sep = "", collapse = ""), "\n")
    return(list(newLL = newLL, model.inf = model.inf, para = para, post.clus=post.clus, Y.hat=Y.sum, score = hu.sum, Z=Z))
  }

